from ..base import BaseModel
from .encoded_text_chunk import EncodedTextChunk


class ModelInput(BaseModel):
    chunks: list[EncodedTextChunk]

    @classmethod
    def from_ints(cls, tokens: list[int]) -> "ModelInput":
        return cls(chunks=[EncodedTextChunk(tokens=tokens)])

    def to_ints(self) -> list[int]:
        return [token for chunk in self.chunks for token in chunk.tokens]

    def tolist(self) -> list[int]:
        return self.to_ints()

    @property
    def length(self) -> int:
        return len(self)

    def __len__(self) -> int:
        return sum([len(chunk) for chunk in self.chunks])



from typing import Any

from pydantic import model_validator

from ..base import BaseModel
from .model_input import ModelInput
from .tensor_data import TensorData


class Datum(BaseModel):
    model_input: ModelInput
    """输入 token ids，为数字列表"""

    loss_fn_inputs: dict[str, TensorData | Any]
    """损失函数，包含的参数有 weights, target_tokens 等"""

    @model_validator(mode="before")
    @classmethod
    def convert_tensordata(cls, data: Any) -> Any:
        if not isinstance(data, dict):
            return data

        loss_fn_inputs = data.get("loss_fn_inputs")
        if not isinstance(loss_fn_inputs, dict):
            return data

        converted = {}
        for key, value in loss_fn_inputs.items():
            converted[key] = TensorData.model_validate(value)

        return {
            **data,
            "loss_fn_inputs": converted,
        }

    def get_data(self, loss_fn: str = "cross_entropy") -> dict:
        """校验参数，并设置为传入给训练服务的参数模型"""
        if loss_fn == "cross_entropy":
            return self._get_data_cross_entropy()
        elif loss_fn == "importance_sampling":
            return self._get_data_importance_sampling()
        elif loss_fn == "ppo":
            return self._get_data_ppo()
        else:
            raise ValueError(f"Unsupported loss function type: {loss_fn}")

    def _get_data_cross_entropy(self):
        if "target_tokens" not in self.loss_fn_inputs:
            raise ValueError("target_tokens is required")
        if len(self.model_input) != len(self.loss_fn_inputs["target_tokens"]):
            raise ValueError(
                f"input tokens ({len(self.model_input)}) and target tokens ({len(self.loss_fn_inputs['target_tokens'])}) must have the same length"
            )
        if "weights" in self.loss_fn_inputs:
            weights = self.loss_fn_inputs["weights"]
        else:
            weights = TensorData(
                data=[1.0] * len(self.model_input),
                dtype="float32",
                shape=[len(self.model_input)],
            )
        return {
            "input_ids": self.model_input.tolist(),
            "target_ids": self.loss_fn_inputs["target_tokens"].tolist(),
            "weights": weights.tolist(),
        }

    def _get_data_importance_sampling(self):
        if "target_tokens" not in self.loss_fn_inputs:
            raise ValueError("target_tokens is required")
        if "logprobs" not in self.loss_fn_inputs:
            raise ValueError("logprobs is required")
        if "advantages" not in self.loss_fn_inputs:
            raise ValueError("advantages is required")
        if (
            len(self.model_input) != len(self.loss_fn_inputs["target_tokens"])
            or len(self.model_input) != len(self.loss_fn_inputs["logprobs"])
            or len(self.model_input) != len(self.loss_fn_inputs["advantages"])
        ):
            raise ValueError(
                "input tokens, target tokens, logprobs, and advantages must have the same length"
            )

        return {
            "input_ids": self.model_input.tolist(),
            "target_ids": self.loss_fn_inputs["target_tokens"].tolist(),
            "logprobs": self.loss_fn_inputs["logprobs"].tolist(),
            "advantages": self.loss_fn_inputs["advantages"].tolist(),
        }

    def _get_data_ppo(self):
        return self._get_data_importance_sampling()



from pathlib import Path

from nanoid import generate


def ensure_file_exists(file_path: str):
    """
    检查文件是否存在，如果不存在则创建一个空文件。
    """
    path = Path(file_path)
    if not path.exists():
        # parents=True: 自动创建不存在的父目录 (如 ./data/config/)
        # exist_ok=True: 防止多线程环境下目录被同时创建导致的冲突
        path.parent.mkdir(parents=True, exist_ok=True)
        # 创建文件
        path.touch(exist_ok=True)


def cuid(size=10):
    """
    生成唯一字符串 cuid，只包含小写字母和数字。默认为 10 位
    """
    return generate("0123456789abcdefghijklmnopqrstuvwxyz", size)


def get_tokenizer(base_model: str):
    try:
        from modelscope import AutoTokenizer
        from modelscope.utils.logger import get_logger
    except ImportError:
        raise ImportError(
            "Detected missing 'modelscope'. This operation requires it. "
            "Install via 'pip install modelscope transformers'"
        )

    logger = get_logger()
    old_level = logger.level

    try:
        logger.setLevel("ERROR")
        return AutoTokenizer.from_pretrained(base_model, local_files_only=True)
    except Exception:
        print(f"Tokenizer not found locally, downloading: {base_model}")
        logger.setLevel("ERROR")
        tokenizer = AutoTokenizer.from_pretrained(base_model, local_files_only=False)
        return tokenizer
    finally:
        logger.setLevel(old_level)



from typing import Literal

from ..base import BaseModel


class EncodedTextChunk(BaseModel):
    tokens: list[int]
    type: Literal["encoded_text"] = "encoded_text"

    def to_ints(self) -> list[int]:
        return list(self.tokens)

    @property
    def length(self) -> int:
        return len(self)

    def __len__(self) -> int:
        return len(self.tokens)



from pathlib import Path
import signal

import click
import tomli as tomllib

from .._utils import PROMPT_PREFIX
from ..config import settings
from ..lib._http import HttpClient


def _load_config(path: Path) -> dict | None:
    if not path.exists():
        return None
    try:
        with open(path, "rb") as f:
            data = tomllib.load(f)
        if not all(k in data for k in ("username", "user_id", "api_key")):
            return None
        return data
    except Exception:
        return None


def _save_config(path: Path, user_info: dict) -> None:
    try:
        path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, "w", encoding="utf-8") as f:
            f.write(f'username = "{user_info["username"]}"\n')
            f.write(f'user_id = "{user_info["user_id"]}"\n')
            f.write(f'api_key = "{user_info["api_key"]}"\n')
            f.write(f'created_at = "{user_info["created_at"]}"\n')
        try:
            path.chmod(0o600)
        except Exception:
            pass
    except Exception as e:
        raise click.ClickException(f"Failed to save config file: {e}") from None


def get_saved_api_key() -> str | None:
    data = _load_config(settings.config_path)
    if not data:
        return None
    api_key = str(data.get("api_key", "")).strip()
    return api_key or None


def _prompt_api_key() -> str:
    click.echo(
        PROMPT_PREFIX
        + "You can find your API key at: "
        + click.style("https://pytrio.cn/dashboard", fg=(229, 229, 18))
    )
    original_sigint_handler = signal.getsignal(signal.SIGINT)
    try:
        signal.signal(signal.SIGINT, signal.default_int_handler)
        try:
            api_key = click.prompt(
                PROMPT_PREFIX
                + "Paste an API key from your profile and hit enter, or press 'Ctrl+C' to quit",
                type=str,
                hide_input=True,
            )
        except KeyboardInterrupt:
            raise click.Abort() from None
    finally:
        signal.signal(signal.SIGINT, original_sigint_handler)

    api_key = str(api_key).strip()
    if not api_key:
        raise click.ClickException(PROMPT_PREFIX + "API key cannot be empty")

    return api_key


def _check_api_key(http_client: HttpClient, api_key: str) -> dict:
    try:
        return http_client.check_user(api_key=api_key)
    except Exception as e:
        raise click.ClickException(str(e)) from None


def _echo_login_success(user_info: dict) -> None:
    click.echo(
        PROMPT_PREFIX
        + "Login successfully. Hi, "
        + click.style(user_info["username"], bold=True)
        + "!"
    )


def ensure_login(
    api_key: str | None = None,
    relogin: bool = False,
    save: bool = True,
    show_success_message: bool = False,
) -> dict:
    """
    进行登录检查
    优先使用传入的 api_key, 其次使用本地保存的 api_key, 如果都没有则手动输入
    """
    config_path = settings.config_path
    http_client = HttpClient()
    prompted_for_api_key = False

    if not relogin and not api_key:
        existing_key = get_saved_api_key()
        if existing_key:
            user_info = _check_api_key(http_client, existing_key)
            if user_info:
                if save:
                    _save_config(config_path, user_info)
                return user_info
            click.echo(PROMPT_PREFIX + "Saved API key is invalid. Please log in again.")

    resolved_api_key = str(api_key).strip() if api_key else _prompt_api_key()
    prompted_for_api_key = not api_key
    if not resolved_api_key:
        raise click.ClickException(PROMPT_PREFIX + "API key cannot be empty")

    user_info = _check_api_key(http_client, resolved_api_key)
    if not user_info:
        click.echo(PROMPT_PREFIX + "Login failed, invalid API key")
        exit(1)

    if save:
        _save_config(config_path, user_info)
    if show_success_message and prompted_for_api_key:
        _echo_login_success(user_info)
    return user_info


@click.option(
    "-k",
    "--api-key",
    type=str,
    help="API Key for authentication",
)
@click.option(
    "-r",
    "--relogin",
    is_flag=True,
    help="Force re-login and overwrite the local API Key",
)
def login(api_key: str | None, relogin: bool):
    """Login to the Trio service"""
    if not relogin and not api_key:
        existing_key = get_saved_api_key()
        if existing_key:
            user_info = _check_api_key(HttpClient(), existing_key)
            if user_info:
                click.echo(
                    PROMPT_PREFIX
                    + "You are already logged in. Use `"
                    + click.style("trio login --relogin", bold=True)
                    + "` to force relogin."
                )
                return

    user_info = ensure_login(api_key=api_key, relogin=relogin, save=True)
    _echo_login_success(user_info)


def logout():
    """Logout from the Trio service"""
    config_path = settings.config_path
    if config_path.exists():
        prompt = (
            PROMPT_PREFIX
            + "Are you sure you want to logout?"
            + click.style(" (y/N):", bold=True)
        )
        if not click.confirm(
            prompt,
            default=False,
            show_default=False,
            prompt_suffix=" ",
        ):
            click.echo(PROMPT_PREFIX + "Logout cancelled.")
            return

        try:
            config_path.unlink()
        except Exception as e:
            raise click.ClickException(f"Failed to delete config file: {e}") from None
        settings.api_key = ""
        click.echo(
            PROMPT_PREFIX
            + "Logout successfully. You can use `"
            + click.style("trio login", bold=True)
            + "` to login again."
        )
        return

    click.echo(
        PROMPT_PREFIX
        + "You are not logged in. If you want to login in, please use `"
        + click.style("trio login", bold=True)
        + "` to login."
    )



from typing import TYPE_CHECKING, Any, Literal

import numpy as np
from pydantic import Field, model_validator

from ..base import BaseModel

if TYPE_CHECKING:
    import torch


class TensorData(BaseModel):
    data: list[float] | list[int]

    dtype: Literal["float32", "int64"]

    shape: list[int] = Field(default_factory=list)

    @model_validator(mode="before")
    @classmethod
    def convert_input(cls, value: Any) -> Any:
        if isinstance(value, cls):
            return value

        try:
            import torch

            if isinstance(value, torch.Tensor):
                return cls.from_torch(value).model_dump()
        except ImportError:
            pass

        if isinstance(value, np.ndarray):
            return cls.from_numpy(value).model_dump()

        if isinstance(value, list):
            arr = np.asarray(value)

            if arr.dtype.kind not in ("i", "u", "f"):
                raise ValueError("Only numeric lists are allowed for TensorData")

            dtype: Literal["float32", "int64"]
            if arr.dtype.kind == "f":
                arr = arr.astype(np.float32)
                dtype = "float32"
            else:
                arr = arr.astype(np.int64)
                dtype = "int64"

            return {
                "data": arr.reshape(-1).tolist(),
                "dtype": dtype,
                "shape": list(arr.shape),
            }

        return value

    @model_validator(mode="after")
    def validate_tensor_data(self) -> "TensorData":
        if not self.shape:
            self.shape = [len(self.data)]

        if any(dim < 0 for dim in self.shape):
            raise ValueError("Negative dimensions are not allowed in shape")

        expected = 1
        for dim in self.shape:
            expected *= dim

        if expected != len(self.data):
            raise ValueError(
                f"Unexpected data length {len(self.data)} for shape {self.shape} (expected {expected})"
            )

        if self.dtype == "float32":
            self.data = [float(x) for x in self.data]
        else:
            normalized: list[int] = []
            for x in self.data:
                if isinstance(x, float) and not x.is_integer():
                    raise ValueError(
                        f"Expected integer values for dtype 'int64', got float {x}"
                    )
                normalized.append(int(x))
            self.data = normalized

        return self

    @classmethod
    def from_numpy(cls, array: np.ndarray) -> "TensorData":
        if np.issubdtype(array.dtype, np.floating):
            array = array.astype(np.float32)
            dtype: Literal["float32", "int64"] = "float32"
        elif np.issubdtype(array.dtype, np.integer):
            array = array.astype(np.int64)
            dtype = "int64"
        else:
            raise ValueError(f"Unsupported numpy dtype for TensorData: {array.dtype}")

        return cls(
            data=array.reshape(-1).tolist(),
            dtype=dtype,
            shape=list(array.shape),
        )

    def to_numpy(self) -> np.ndarray:
        if self.dtype == "float32":
            numpy_dtype = np.float32
        elif self.dtype == "int64":
            numpy_dtype = np.int64
        else:
            raise ValueError(f"Unsupported TensorDtype: {self.dtype}")

        arr = np.array(self.data, dtype=numpy_dtype)
        if self.shape:
            arr = arr.reshape(self.shape)
        return arr

    @classmethod
    def from_torch(cls, tensor: "torch.Tensor") -> "TensorData":
        try:
            import torch
        except ImportError as e:
            raise ImportError(
                "Detected missing 'torch'. This operation requires it. "
                f"Install via 'pip install torch'. Error details: {e}"
            )

        if tensor.dtype.is_floating_point:
            tensor = tensor.to(torch.float32)
            dtype: Literal["float32", "int64"] = "float32"
        else:
            tensor = tensor.to(torch.int64)
            dtype = "int64"

        return cls(
            data=tensor.detach().cpu().reshape(-1).tolist(),
            dtype=dtype,
            shape=list(tensor.shape),
        )

    def tolist(self) -> list[Any]:
        return self.to_numpy().tolist()

    def __len__(self) -> int:
        return len(self.data)
